// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Kermin Fleming kfleming@mit.edu

import Connectable::*;
import GetPut::*;
import ClientServer::*;
import FIFO::*;
import SpecialFIFOs::*;
import Vector::*;

//AWB includes
`include "asim/provides/low_level_platform_interface.bsh"
`include "asim/provides/soft_connections.bsh"
`include "asim/provides/common_services.bsh"

//Local includes
`include "asim/provides/audio_pipe_types.bsh"  //provides Audio Pipeline interface
`include "asim/provides/path_types.bsh"
`include "asim/provides/core.bsh"
`include "asim/provides/mixer.bsh"
`include "asim/provides/processor_library.bsh"
`include "asim/provides/fpga_components.bsh"
`include "asim/dict/VDEV_SCRATCH.bsh"

`include "asim/rrr/remote_client_stub_AUDIOCORERRR.bsh"
`include "asim/rrr/remote_server_stub_AUDIOCORERRR.bsh"

`define MAX_VOICES       2

module [CONNECTED_MODULE] mkConnectedApplication ();

   Vector#(`MAX_VOICES, Volume) channelVols = replicate(127);
   Mixer mixer <- mkMixer(`MAX_VOICES, channelVols); //should be max voices but 2 for now

   Reg#(Bit#(32)) cycle <- mkReg(0);
   Reg#(Bit#(32)) sampleCountIn <-mkReg(0);
   Reg#(Bit#(32)) sampleCountOut <-mkReg(0);
   Vector#(`MAX_VOICES, Reg#(Bit#(32))) pc <-replicateM(mkReg(0)); //tracks pcs from cores for debug purposes
   Vector#(`MAX_VOICES, Reg#(Bool)) channelFini <- replicateM(mkReg(False));

   // Services Samples 
   ClientStub_AUDIOCORERRR client_stub <- mkClientStub_AUDIOCORERRR();   
  
   //-----------------------------------------------------------
   // Debug port
   
   ServerStub_AUDIOCORERRR server_stub <- mkServerStub_AUDIOCORERRR();   

   // Create Cores
   Vector#(`MAX_VOICES, Core) cores;
   for (Integer n = 0; n < `MAX_VOICES; n = n + 1)
     begin
	 case (n)
	    0 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYA);
	    1 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYB);
	    2 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYC);
	    3 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYD);
	    4 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYE);
	    5 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYF);
	    6 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYG);
	    7 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYH);
	    8 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYI);
	    9 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYJ);
	    10 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYK);
	    11 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYL);
	    12 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYM);
	    13 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYN);
	    14 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYO);
	    15 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYP);
	    16 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYQ);
	    17 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYR);
	    18 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYS);
	    19 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYT);
	    20 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYU);
	    21 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYV);
	    22 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYW);
	    23 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYX);
	    24 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYY);
	    25 : cores[n] <- mkCore(`VDEV_SCRATCH_MEMORYZ);
	 endcase
     end

   function Integer countFiniVoices();
      Integer count = 0;
      for (Integer i = 0; i < `MAX_VOICES; i = i + 1)
	if (channelFini[i]) count = count + 1;
      return count;
   endfunction
      
  // this is for the tracing
  rule printCycles;
    cycle <= cycle+1;
    $fdisplay(stderr, " => Cycle = %d", cycle);
  endrule

  // get the pc for trace purposes
   rule getPC;
     for (Integer i = 0; i < `MAX_VOICES; i = i + 1)
	begin
	   let val <- cores[i].pc.get(); 
	   pc[i] <= val;
	end 
   endrule

  // Send to Mixer
  // Right now this is sorta retarded in that I pass from the output fifo into a new fifo
  // But I have to mod a bunch of things to fix this and I am not sure I understand
  // things well enough to do this quickly.  So here it is as it is for now.
   rule mix;
      for (Integer i = 0; i < `MAX_VOICES; i = i + 1)
        begin
	   let coreOut <- cores[i].sampleOutput.get();
	   mixer.toMixer(AudioStream {voice : fromInteger(i), data:coreOut.data});
	end
   endrule
   
  rule feedOutput;
     let pipeOut <- mixer.mainOut.get();
     
     AudioProcessorControl endOfFileTag = EndOfFile;
     AudioProcessorControl sampleTag = Data;

     sampleCountOut <= sampleCountOut+1;
     
     $display("PIPE writes sample %d", sampleCountOut);
     case (pipeOut) matches
	tagged EndOfFile:
	   begin
	      client_stub.makeRequest_SendProcessedStream(zeroExtend(pack(endOfFileTag)),?);
	      for (Integer i = 1; i < `MAX_VOICES; i = i+1) channelFini[i] <= False; 
	            //This will report conflict with
		    //the other setting of channelFini at input except they are mutually
		    //exclusive since this is only hit after all input finished
	   end
       tagged Sample .sample:
	  client_stub.makeRequest_SendProcessedStream(zeroExtend(pack(sampleTag)), zeroExtend(pack(sample)));
     endcase
  endrule

   // Can generally just stick with the EOF but since I have two Cores no mixer...
  /*(* conservative_implicit_conditions *)
   rule sendTerminate;
      Bool done = True;
      for (Integer i = 1; i < 2; i = i+1)
	 done = ac_fini[i] && done;

      if (done)
	 client_stub.makeRequest_SendTerminate(zeroExtend(pack(1)));
      
   endrule
*/

   //*****  SERVER Side  *****
  // this requires me to switch the processor to handle this which I will do later.
   // When a file has finished, want the system to pass invalid audiostreams
   rule fillFinished;
      for (Integer i = 1; i < `MAX_VOICES; i = i+1)
	 begin
	    if (channelFini[i])
	       cores[i].sampleInput.put(AudioStream {voice : fromInteger(i), data: tagged Invalid});
	 end
   endrule
   
   (* conservative_implicit_conditions *)
   rule feedInput;
     let command <- server_stub.acceptRequest_SendUnprocessedStream();
     AudioProcessorControl ctrl = unpack(truncate(command.ctrl));

     VoiceId channel = unpack(truncate(command.channel));
//     $display("PIPE: Incoming sample to channel %x", channel);
	   
     AudioProcessorUnit inSample;
      
     if(ctrl == EndOfFile)
       begin
	   $display("lsp: PIPE received EOF   ");
	  inSample = tagged EndOfFile;
          channelFini[channel] <= True;
       end
     else 
	begin
//	   $display("lsp: PIPE received Data   ");
           inSample = tagged Sample unpack(truncate(command.sample));
	end

      if (!channelFini[channel])
	 cores[channel].sampleInput.put(AudioStream {voice: channel, data: tagged Valid inSample });
      else $display("PIPE ERROR: Trying to send to Stream officially done");
      
   endrule

   (* conservative_implicit_conditions *)
   rule handleCPUToHost;
      let req <- server_stub.acceptRequest_ReadCPUToHost();
      case (req) 
	 1: server_stub.sendResponse_ReadCPUToHost(cycle);
	 2: server_stub.sendResponse_ReadCPUToHost(sampleCountIn);
	 3: server_stub.sendResponse_ReadCPUToHost(sampleCountOut);
	 4: server_stub.sendResponse_ReadCPUToHost(fromInteger(countFiniVoices()));
	 default : if (req < 10 + `MAX_VOICES) 
		      server_stub.sendResponse_ReadCPUToHost(pc[req-10]);  
      endcase
  endrule

endmodule
